#pragma once;

#include "Orbitersdk.h"
#include "XRVesselCtrl.h"
#include "Log.h"

#define MSG_MAXSTRLEN 32

#define MAX_FUEL_TANKS 16
#define MAX_THRUSTERS 64
#define MAX_DOCKING_PORTS 16
#define MAX_NAV_RADIOS 16
#define MAX_ANIMATIONS 16

#define MESSAGE_TYPE_START		((char) 0)
#define MESSAGE_TYPE_END		((char) 255)
#define MESSAGE_TYPE_VESSEL		((char) 1)
#define MESSAGE_TYPE_VESSEL2	((char) 2)
#define MESSAGE_TYPE_AIRFOIL	((char) 3)
#define MESSAGE_TYPE_XR			((char) 4)
#define MESSAGE_TYPE_FOCUS		((char) 5)
#define MESSAGE_TYPE_VESSELSTATE ((char) 6)
#define MESSAGE_TYPE_KEYBOARD	((char) 7)
#define MESSAGE_TYPE_VESSELEXTRA ((char) 8)
#define MESSAGE_TYPE_ANIMATION	 ((char) 9)
#define MESSAGE_TYPE_ACK		((char) 250)
#define MESSAGE_TYPE_HELLO		((char) 251)

#define MESSAGE_ERROR			((Message*) 0x0)
#define MESSAGE_TIMED_OUT		((Message*) 0x1)
#define MESSAGE_VALID(x)		( x > (Message*) 0x10 ) 

typedef char text_t[MSG_MAXSTRLEN];

typedef struct _VesselMessage {
	double mjd;
	text_t name;
	text_t classname;
	VESSELSTATUS status;
	text_t rbody;
	text_t base;
} VesselMessage_t;

typedef struct _VesselExtraMessage {
	text_t vesselName;
	int navModeState;
	int rcsMode;
	int adCtrlMode;
	WORD navChannel[MAX_NAV_RADIOS]; 
} VesselExtraMessage_t;

typedef struct _AirfoilMessage {
	text_t vesselName;
	double elevator;
	double rudder;
	double aileron;
	double flap;
	double elevatorTrim;
	double rudderTrim;
} AirfoilMessage_t;

typedef struct _AnimationMessage {
	text_t vesselName;
	int nAnimations;
	double state[MAX_ANIMATIONS];
} AnimationMessage_t;

typedef struct _Vessel2Message {
	double mjd;
	text_t name;
	text_t classname;
	VESSELSTATUS2 status;
	text_t rbody;
	text_t base;
	VESSELSTATUS2::FUELSPEC fuelspec[MAX_FUEL_TANKS];
	VESSELSTATUS2::THRUSTSPEC thrustspec[MAX_THRUSTERS];
	VESSELSTATUS2::DOCKINFOSPEC dockinfo[MAX_DOCKING_PORTS];
	text_t dockedVessel[MAX_DOCKING_PORTS];
} Vessel2Message_t;

typedef struct _VesselStateMessage {
	struct _subhdr {
		double mjd;
		text_t name;
		int length;
	} subheader;
	char buf[65536];
} VesselStateMessage_t;

typedef struct _keyevent {
	int ncode;
	WPARAM wparam;
	LPARAM lparam;
} keyevent_t;

#define MAX_KEYBOARD_EVENTS 16
typedef struct _KeyboardMessage {
	int len;
	keyevent_t events[MAX_KEYBOARD_EVENTS];
} KeyboardMessage_t;

#define XR_DOORS		15
#define XR_LIGHTS		3
#define XR_ENGINES		8

typedef struct _XRMessage {
	text_t vesselName;
	XRSystemStatusWrite xrSystemStatus;
	bool xrLightsStatus[XR_LIGHTS];
	int xrDoorStatus[XR_DOORS];
	double xrCenterOfGravity;
	bool xrRCSDockingMode;
	bool xrElevatorEVAPortActive;
	XREngineStateWrite xrEngineStatus[XR_ENGINES];
} XRMessage_t;

typedef struct _FocusMessage {
	text_t vesselName;
} FocusMessage_t;

typedef struct _StartMessage {
	double mjd;
	double timeAccel;
	double simStep;
} StartMessage_t;

typedef struct _MessageHeader {
	int type;
	int serial;
} MessageHeader_t;

typedef struct _Message {
	MessageHeader_t header;
	union {
		StartMessage_t start;
		VesselMessage_t vessel;
		Vessel2Message_t vessel2;
		VesselStateMessage_t vesselstate;
		VesselExtraMessage_t vesselextra;
		KeyboardMessage_t keyboard;
		AirfoilMessage_t airfoil;
		AnimationMessage_t animation;
		XRMessage_t xr;
		FocusMessage_t focus;
	} msg; 
} Message_t;


void DumpMessage(Message_t *msg);

class Message {
protected:
	Message_t buffer;

public:
	Message(char type);
	Message(Message_t *msg);
	virtual int GetSize(void);
	char GetType(void) { return buffer.header.type; };
	int GetSerial(void) { return buffer.header.serial; };
	int Send(SOCKET s);
	int Send(SOCKET s, struct sockaddr *client);
	void Send(SOCKET s, vector<struct sockaddr> *clients);
	virtual ~Message(void) { };
};

class EndMessage : public Message 
{
public:
	EndMessage(Message_t *msg) : Message(msg) { };
	EndMessage(void) : Message(MESSAGE_TYPE_END) {	};
};


class StartMessage : public Message 
{
public:
	StartMessage(Message_t *msg) : Message(msg) { };
	StartMessage(void) : Message(MESSAGE_TYPE_START) {
		buffer.msg.start.mjd = oapiGetSimMJD();
		buffer.msg.start.simStep = oapiGetSimStep();
		buffer.msg.start.timeAccel = oapiGetTimeAcceleration();
	};
	double GetMJD(void) { return buffer.msg.start.mjd; };
	double GetTimeAcceleration(void) { return buffer.msg.start.timeAccel; };
	double GetSimStep(void) { return buffer.msg.start.simStep; };
};

class VesselMessage : public Message
{
public:
	//VesselMessage(void) : Message(MESSAGE_TYPE_START) { };
	VesselMessage(OBJHANDLE vessel);
	VesselMessage(Message_t *msg);
	VESSELSTATUS *GetStatus(void) { return &buffer.msg.vessel.status; };
	char *GetName(void) { return buffer.msg.vessel.name; };
	char *GetClassName(void) { return buffer.msg.vessel.classname; };
	void UpdateVessel(void);
};

class Vessel2Message : public Message
{
public:
	//Vessel2Message(void) : Message(MESSAGE_TYPE_START) { };
	Vessel2Message(OBJHANDLE vessel);
	Vessel2Message(Message_t *msg);
	VESSELSTATUS2 *GetStatus(void) { return &buffer.msg.vessel2.status; };
	char *GetName(void) { return buffer.msg.vessel2.name; };
	char *GetClassName(void) { return buffer.msg.vessel2.classname; };
	void UpdateVessel(void);
};

class VesselStateMessage : public Message
{
public:
	VesselStateMessage(OBJHANDLE vessel);
	VesselStateMessage(Message_t *msg) : Message(msg) { } ;
	void UpdateVessel(void);
	char *GetName(void) { return buffer.msg.vesselstate.subheader.name; };
};

class AirfoilMessage : public Message
{
public:
	AirfoilMessage(OBJHANDLE vessel);
	AirfoilMessage(Message_t *msg) : Message(msg) { } ;
	void UpdateVessel(void);
	char *GetName(void) { return buffer.msg.airfoil.vesselName; };
};

class AnimationMessage : public Message
{
public:
	AnimationMessage(OBJHANDLE vessel);
	AnimationMessage(Message_t *msg) : Message(msg) { } ;
	void UpdateVessel(void);
	char *GetName(void) { return buffer.msg.animation.vesselName; };
};

class VesselExtraMessage : public Message
{
public:
	VesselExtraMessage(OBJHANDLE vessel);
	VesselExtraMessage(Message_t *msg) : Message(msg) { } ;
	void UpdateVessel(void);
	char *GetName(void) { return buffer.msg.vesselextra.vesselName; };
};

class XRMessage : public Message
{
	bool bValid;
public:
	XRMessage(OBJHANDLE vessel);
	XRMessage(Message_t *msg) : Message(msg) { bValid = TRUE; } ;
	void UpdateVessel(void);
	char *GetName(void) { return buffer.msg.xr.vesselName; };
	bool IsValid(void) { return bValid; };
};

class FocusMessage : public Message
{
public:
	FocusMessage(void);
	FocusMessage(Message_t *msg) : Message(msg) { } ;
	void SwitchFocus(void);
};

class KeyboardMessage : public Message
{
public:
	KeyboardMessage(void);
	KeyboardMessage(Message_t *msg) : Message(msg) { } ;
	void HookProc(int ncode, WPARAM wparam, LPARAM lparam);
	void SetState(void);
	void Clear(void);
};

class AckMessage : public Message
{
public:
	AckMessage(void) : Message(MESSAGE_TYPE_ACK) { } ;
	AckMessage(Message_t *msg) : Message(msg) { } ;
};

class HelloMessage : public Message
{
public:
	HelloMessage(void) : Message(MESSAGE_TYPE_HELLO) { } ;
	HelloMessage(Message_t *msg) : Message(msg) { } ;
};

u_long inline MessageSocketHasData(SOCKET sock)
{
	u_long bytes;
	ioctlsocket(sock, FIONREAD, &bytes);
	return bytes;
}


Message *CreateVesselMessage(OBJHANDLE vessel);
Message *CreateVesselStateMessage(OBJHANDLE vessel);

Message *ParseMessage(Message_t *msg);
Message *ReceiveMessage(SOCKET s, sockaddr *from, int *fromlen);
Message *ReceiveMessage(SOCKET s);
Message *ReceiveMessage(SOCKET s, int timeOutMs);


extern KeyboardMessage KeyboardMessageBuf;
void HookKeyboard(char *moduleName);
void UnHookKeyboard(void);
